@@ -1,5 +1,6 @@
 from .elasticsearch_operator import index_json_bulk, index_json, index_json_bulk_parallel, ElasticsearchOperator
-from .redshift_operator import RedshiftOperator, send_to_redshift, read_from_redshift
+from .redshift_operator import RedshiftOperator, send_to_redshift, read_from_redshift, send_metadata_to_redshift, \
+    find_tables_by_column_name
 from .snowflake_operator import SnowflakeOperator, migrate_metadata_from_redshift
 from .s3_operator import S3Operator, upload_raw_s3, download_raw_s3
 from .gcloud_operator import upload_to_gcloud, download_from_gcloud, GcloudOperator
