@@ -1,58 +1,91 @@
-import os
-import re
 from typing import List, Optional
-
+from .redshift_operator import RedshiftOperator
+from .s3_operator import S3Operator
+from ..sql_utils import convert_to_snowflake_syntax, correct_sql_system_variables_syntax
 import pandas as pd
 import snowflake.connector
 import sqlparse
-
-from .redshift_operator import RedshiftOperator
-from .s3_operator import S3Operator
-from ..sql_utils import convert_to_snowflake_syntax
+import os
+import re
 
 
-def migrate_data_from_redshift(prefix: str, exclude_tables: List[str], bucket: Optional[str] = None,  ):
-    bucket_s3 = (os.environ.get("S3_BUCKET_DATAMART") if bucket is None else bucket)
-
+def migrate_data_from_redshift(rs_source_schema: str,
+                               sn_target_schema: str,
+                               exclude_tables: List[str],
+                               bucket: Optional[str]):
+    """
+    If we want to migrate the data itself from Redshift into Snowflake, there is a SQL method on Snowflake
+    to copy the tables directly from S3 into Snowflake.
+    :param rs_source_schema(str): Source schema name in Redshift.
+    :param sn_target_schema(str): Source schema name in Snowflake.
+    :param exclude_tables(List[str]): Tables to be excluded when migrating the data.
+    :param bucket(Optional[str]): Bucket name in S3. If none is passed, it will take the bucket name from
+                                  the environment variable S3_BUCKET_DATAMART.
+    :return:
+    """
     so = S3Operator()
-    so.bucket = bucket_s3
-    so.prefix = prefix
+    so.bucket = os.environ.get("S3_BUCKET_DATAMART") if bucket is None else bucket
+    so.prefix = f"{rs_source_schema}/"
 
     if so.detect_prefix():
         list_obj = so.list_s3_obj()
         list_tables = [obj.split("/")[-1].split(".")[0] for obj in list_obj if
                        ((len(obj.split("/")[-1]) > 0) and (obj.split(".")[-1] == "csv"))]
     else:
-        raise ValueError(f'❌ The chosen prefix {prefix} does not exist! Make sure to have a slash at the end of the '
-                         f'name: <prefix_name>/')
+        raise ValueError(f'❌ The chosen schema {so.prefix} does not exist in S3!')
 
-    list_tables = list(set(list_tables).difference(exclude_tables))
+    list_tables = list(set(list_tables).difference(exclude_tables)) if exclude_tables is not None else list_tables
 
+    # We first migrate the metadata in order to create the tables with all the columns and its descriptions.
+    # Only after we are sure all the tables were created in snowflake, we can migrate the data itself from s3
+    # to the snowflake tables.
     migrate_metadata_from_redshift(rs_db="oip",
                                    sf_db="oip",
-                                   schemas_list=["pv_intermediary_tables"],
+                                   schemas_list=[rs_source_schema],
                                    create_tables=True,
                                    tables_list=list_tables,
                                    not_include=[],
-                                   schema_migration_dict={"pv_intermediary_tables": "emea_sales"})
+                                   schema_migration_dict={rs_source_schema: sn_target_schema})
 
     sf = SnowflakeOperator()
     for table in list_tables:
-        sf.copy_from_S3("pv_intermediary_tables", f"{table}.csv", "OIP", "emea_sales".upper(), table)
-    print("finished")
+        sf.copy_from_S3(rs_source_schema, f"{table}.csv", "oip", "emea_sales", table)
+
+    sf.conn.cursor().close()
+    sf.conn.close()
+    print("✅ Process finished successfully!")
 
 
-def migrate_metadata_from_redshift(
-        rs_db: str,
-        sf_db: str,
-        schemas_list: List = [],
-        tables_list: List = [],
-        create_tables: bool = False,
-        logging: bool = True,
-        verbose: bool = True,
-        not_include: Optional = None,
-        schema_migration_dict: Optional[dict] = None
-):
+def migrate_metadata_from_redshift(rs_db: str,
+                                   sf_db: str,
+                                   schemas_list: List[str] = [],
+                                   tables_list: List[str] = [],
+                                   create_tables: bool = False,
+                                   logging: bool = True,
+                                   verbose: bool = True,
+                                   not_include: Optional[List[str]] = None,
+                                   schema_migration_dict: Optional[dict] = None):
+    """
+    Airbyte is our main tool for Data Migration. However, it is responsible only for the data migration, not metadata.
+    This function aims, therefore, to retrieve all the DDL Queries on Redshift and replicate then in Snowflake.
+    :param rs_db(str): Redshift database name.
+    :param sf_db(str): Snowflake database name.
+    :param schemas_list(List[str]): List of schemas to be migrated.
+    :param tables_list(List[str]): List of tables to be migrated.
+    :param create_tables(bool): If the tables being migrated don't exist beforehand, they must be created.
+    :param logging(bool): Since SQL Syntax errors can be pretty hard to understand and debug, we can set logging
+                          to True in order to have the problematic queries written down into the query_log_errors.txt
+                          file.
+    :param verbose(bool): See all the DDL Queries being read from the redshift tables.
+    :param not_include(Optional[List[str]]): List of schemas not to be included.
+    :param schema_migration_dict(Optional[dict]): If the migration happens to be between schemas with different names,
+                                                  we can set this config as a dictionary.
+                                                  - For example:
+                                                        If we want to migrate the metadata from schema1 in Redshift
+                                                        into schema2 in Snowflake, we can set this relation as:
+                                                        >>> schema_migration_dict = {schema1: schema2}.
+    :return:
+    """
     print("🦆 Starting the metadata migration process | Redshift -> Snowflake")
     ro = RedshiftOperator(database=rs_db)
     ddl_model = ro.get_DDL(verbose=verbose,
@@ -82,27 +115,14 @@ def migrate_metadata_from_redshift(
         # Some of these corrections below must be done because 'year', 'level', 'region', 'names' are SQL Syntax Names
         # and AWS parse them as strings when creating the column names. However, Snowflake parses it otherwise because
         # it can distinguish the column names and the SQL Variables as different things.
-        ddl_model.create_query = ddl_model.create_query.str.replace('"year"', "year ").str.replace('"level" ', "level ")\
-            .str.replace('"region" ', "region ")\
-            .str.replace('"names" ', "names ")\
-            .str.replace('"type" ', "type ")\
-            .str.replace('"role" ', "role ")\
-            .str.replace('"provider" ', "provider ")\
-            .str.replace('"location" ', "location ")
+        ddl_model["create_query"] = correct_sql_system_variables_syntax(ddl_model, "create_query")
         sf.execute_metadata_query(ddl_model.create_query.values, logging=logging, correct_syntax=True)
 
     print("🏷 Migrating Table Descriptions...")
     sf.execute_metadata_query(ddl_model.table_description.values, logging=logging)
 
     print("🏷 Migrating Columns Descriptions...")
-    ddl_model.columns_description = ddl_model.columns_description.str.replace('."year"', ".year")\
-        .str.replace('."level"', ".level")\
-        .str.replace('."region"', ".region")\
-        .str.replace('."names"', ".names")\
-        .str.replace('."type"', ".type")\
-        .str.replace('."role"', ".role")\
-        .str.replace('."provider"', ".provider")\
-        .str.replace('."location"', ".location ")
+    ddl_model["columns_description"] = correct_sql_system_variables_syntax(ddl_model, "columns_description")
     sf.execute_metadata_query(ddl_model.columns_description.values, logging=logging)
 
     print("🔑 Migrating Primary Keys...")
@@ -119,30 +139,14 @@ def migrate_metadata_from_redshift(
     print("✅ All metadata have been migrated successfully !")
 
 
-# def update_to_snowflake(self):
-#     snow = SnowflakeOperator()
-#     print("process prices")
-#     snow.conn.cursor().execute(f"""    delete from OIP_PREMIUM.PRICE_TRACKER.drug_prices where (global_drug_id LIKE '{self.country_code}%');
-#     """)
-#     snow.copy_from_S3(s3_prefix=os.path.join("udm", "price_tracker"), s3_file=f"{self.country_code}__drug_prices",
-#               database="OIP_PREMIUM", schema="PRICE_TRACKER", table="drug_prices")
-#     print("process products")
-#     snow.conn.cursor().execute(f"""    delete from OIP_PREMIUM.PRICE_TRACKER.drug_products where (global_drug_id LIKE '{self.country_code}%');
-#     """)
-#     snow.copy_from_S3(s3_prefix=os.path.join("udm", "price_tracker"), s3_file=f"{self.country_code}__drug_products",
-#               database="OIP_PREMIUM", schema="PRICE_TRACKER", table="drug_products")
-
-
 class SnowflakeOperator:
-    def __init__(
-            self,
-            snowflake_user=None,
-            snowflake_password=None,
-            snowflake_account=None,
-            snowflake_warehouse=None,
-            snowflake_role=None,
-            snowflake_database=None,
-    ) -> None:
+    def __init__(self,
+                 snowflake_user=None,
+                 snowflake_password=None,
+                 snowflake_account=None,
+                 snowflake_warehouse=None,
+                 snowflake_role=None,
+                 snowflake_database=None) -> None:
 
         self.snowflake_user = (
             os.environ.get("SNOWFLAKE_USER")
@@ -189,10 +193,20 @@ class SnowflakeOperator:
         self.conn.cursor().execute(sql)
 
     def copy_from_S3(self, s3_prefix, s3_file, database, schema, table):
+        """
+        The Snowflake SGBD has a SQL Command dedicated to copy tables directly from S3 into a schema.
+
+        :param s3_prefix: Object prefix in s3.
+        :param s3_file: Table name in s3.
+        :param database: Database name in Snowflake.
+        :param schema: Schema name in Snowflake.
+        :param table: Table name in Snowflake.
+        :return:
+        """
         self.conn.cursor().execute(
             f"""
-            COPY INTO {database}.{schema}.{table} 
-            FROM s3://{os.environ.get("S3_DATAMART_BUCKET")}/{s3_prefix}/{s3_file}
+            COPY INTO {database.upper()}.{schema.upper()}.{table.upper()} 
+            FROM s3://{os.environ.get("S3_DATAMART_BUCKET")}/{s3_prefix.lower()}/{s3_file.lower()}
             CREDENTIALS = (
             aws_key_id='{os.environ.get("AWS_ACCESS_KEY_ID")}',
             aws_secret_key='{os.environ.get("AWS_SECRET_ACCESS_KEY")}'
@@ -203,23 +217,27 @@ class SnowflakeOperator:
             """
         )
 
-    # def get_information_schema(self):
-    #     self.conn.cursor().execute(f"{command.strip()};")
-
     def correct_syntax(self, query: str, no_comments: bool = False) -> str:
         """
-        Also, snowflake has a specific SQL syntax. If we get a SQL from another SGBD, we must firstly
+        Snowflake has a specific SQL syntax. If we get a SQL from another SGBD, we must firstly
         correct the syntax into the snowflake constraints.
 
         Args:
             query: String containing the SQL script.
-            no_comments:
+            no_comments: If we want to keep the SQL comments as well or not. Recommended keeping as false.
 
         Returns: The same query but snowflake-compatible.
         """
         return convert_to_snowflake_syntax(query, no_comments)
 
     def execute_metadata_query(self, query: List[str], logging: bool = False, correct_syntax=False):
+        """
+
+        :param query:
+        :param logging:
+        :param correct_syntax:
+        :return:
+        """
         # The snowflake python API allows only one command per request. That's why we must split the input query into
         # a list of commands.
         # Also, snowflake has a specific SQL syntax. If we get a SQL from another database manager, we must firstly
@@ -245,7 +263,7 @@ class SnowflakeOperator:
                         print(f"Problem found. Skipping command...")
 
     def execute_key_query(self, df: pd.DataFrame, key: str = "primary", logging: bool = False):
-        for row in range(len(df)):
+        for row in range(df.__len__()):
             if key == "primary":
                 if df.iloc[row, 2] != "":
                     drop_key_query = f"ALTER TABLE {df.iloc[row, 0]}.{df.iloc[row, 1]} DROP PRIMARY KEY;"
